@@ -777,6 +777,27 @@ Address=1.2.3.4/12
 unmanaged-devices+=interface-name:br0,''')
         self.assert_udev(None)
 
+    def test_bridge_forward_declaration(self):
+        self.generate('''network:
+  version: 2
+  bridges:
+    br0:
+      interfaces: [eno1, switchports]
+      dhcp4: true
+  ethernets:
+    eno1: {}
+    switchports:
+      match:
+        driver: yayroute
+''')
+
+        self.assert_networkd({'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n',
+                              'br0.network': ND_DHCP4 % 'br0',
+                              'eno1.network': '[Match]\nName=eno1\n\n'
+                                              '[Network]\nBridge=br0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n',
+                              'switchports.network': '[Match]\nDriver=yayroute\n\n'
+                                                     '[Network]\nBridge=br0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n'})
+
     def test_bridge_components(self):
         self.generate('''network:
   version: 2
@@ -797,6 +818,42 @@ unmanaged-devices+=interface-name:br0,''')
                               'switchports.network': '[Match]\nDriver=yayroute\n\n'
                                                      '[Network]\nBridge=br0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n'})
 
+    def test_bridge_params(self):
+        self.generate('''network:
+  version: 2
+  ethernets:
+    eno1: {}
+    switchports:
+      match:
+        driver: yayroute
+  bridges:
+    br0:
+      interfaces: [eno1, switchports]
+      parameters:
+        ageing-time: 50
+        priority: 1000
+        forward-delay: 12
+        hello-time: 6
+        max-age: 24
+        stp: true
+        path-cost:
+          eno1: 70
+      dhcp4: true''')
+
+        self.assert_networkd({'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n\n'
+                                            '[Bridge]\nAgeingTimeSec=50\n'
+                                            'Priority=1000\n'
+                                            'ForwardDelaySec=12\n'
+                                            'HelloTimeSec=6\n'
+                                            'MaxAgeSec=24\n'
+                                            'STP=true\n',
+                              'br0.network': ND_DHCP4 % 'br0',
+                              'eno1.network': '[Match]\nName=eno1\n\n'
+                                              '[Network]\nBridge=br0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n\n'
+                                              '[Bridge]\nCost=70\n',
+                              'switchports.network': '[Match]\nDriver=yayroute\n\n'
+                                                     '[Network]\nBridge=br0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n'})
+
     def test_bond_empty(self):
         self.generate('''network:
   version: 2
@@ -831,6 +888,89 @@ unmanaged-devices+=interface-name:bn0,''')
                               'switchports.network': '[Match]\nDriver=yayroute\n\n'
                                                      '[Network]\nBond=bn0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n'})
 
+    def test_bond_empty_parameters(self):
+        self.generate('''network:
+  version: 2
+  ethernets:
+    eno1: {}
+    switchports:
+      match:
+        driver: yayroute
+  bonds:
+    bn0:
+      parameters: {}
+      interfaces: [eno1, switchports]
+      dhcp4: true''')
+
+        self.assert_networkd({'bn0.netdev': '[NetDev]\nName=bn0\nKind=bond\n',
+                              'bn0.network': ND_DHCP4 % 'bn0',
+                              'eno1.network': '[Match]\nName=eno1\n\n'
+                                              '[Network]\nBond=bn0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n',
+                              'switchports.network': '[Match]\nDriver=yayroute\n\n'
+                                                     '[Network]\nBond=bn0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n'})
+
+    def test_bond_with_parameters(self):
+        self.generate('''network:
+  version: 2
+  ethernets:
+    eno1: {}
+    switchports:
+      match:
+        driver: yayroute
+  bonds:
+    bn0:
+      parameters:
+        mode: 802.1ad
+        lacp-rate: 10
+        mii-monitor-interval: 10
+        min-links: 10
+        up-delay: 10
+        down-delay: 10
+        all-slaves-active: true
+        transmit-hash-policy: none
+        ad-select: none
+        arp-interval: 10
+        arp-validate: all
+        arp-all-targets: all
+        fail-over-mac-policy: none
+        gratuitious-arp: 10
+        packets-per-slave: 10
+        primary-reselect-policy: none
+        resend-igmp: 10
+        learn-packet-interval: 10
+        arp-ip-targets:
+          - 10.10.10.10
+          - 20.20.20.20
+      interfaces: [eno1, switchports]
+      dhcp4: true''')
+
+        self.assert_networkd({'bn0.netdev': '[NetDev]\nName=bn0\nKind=bond\n\n'
+                                            '[Bond]\n'
+                                            'Mode=802.1ad\n'
+                                            'LACPTransmitRate=10\n'
+                                            'MIIMonitorSec=10\n'
+                                            'MinLinks=10\n'
+                                            'TransmitHashPolicy=none\n'
+                                            'AdSelect=none\n'
+                                            'AllSlavesActive=1\n'
+                                            'ARPIntervalSec=10\n'
+                                            'ARPIPTargets=10.10.10.10,20.20.20.20\n'
+                                            'ARPValidate=all\n'
+                                            'ARPAllTargets=all\n'
+                                            'UpDelaySec=10\n'
+                                            'DownDelaySec=10\n'
+                                            'FailOverMACPolicy=none\n'
+                                            'GratuitiousARP=10\n'
+                                            'PacketsPerSlave=10\n'
+                                            'PrimaryReselectPolicy=none\n'
+                                            'ResendIGMP=10\n'
+                                            'LearnPacketIntervalSec=10\n',
+                              'bn0.network': ND_DHCP4 % 'bn0',
+                              'eno1.network': '[Match]\nName=eno1\n\n'
+                                              '[Network]\nBond=bn0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n',
+                              'switchports.network': '[Match]\nDriver=yayroute\n\n'
+                                                     '[Network]\nBond=bn0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n'})
+
     def test_gateway(self):
         self.generate('''network:
   version: 2
@@ -1715,6 +1855,58 @@ interface-name=br0
 [ipv4]
 method=auto
 address1=1.2.3.4/12
+'''})
+        self.assert_networkd({})
+        self.assert_udev(None)
+
+    def test_bridge_forward_declaration(self):
+        self.generate('''network:
+  version: 2
+  renderer: NetworkManager
+  bridges:
+    br0:
+      interfaces: [eno1, switchport]
+      dhcp4: true
+  ethernets:
+    eno1: {}
+    switchport:
+      match:
+        name: enp2s1
+''')
+
+        self.assert_nm({'eno1': '''[connection]
+id=netplan-eno1
+type=ethernet
+interface-name=eno1
+slave-type=bridge
+master=br0
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+''',
+                        'switchport': '''[connection]
+id=netplan-switchport
+type=ethernet
+interface-name=enp2s1
+slave-type=bridge
+master=br0
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+''',
+                        'br0': '''[connection]
+id=netplan-br0
+type=bridge
+interface-name=br0
+
+[ipv4]
+method=auto
 '''})
         self.assert_networkd({})
         self.assert_udev(None)
@@ -1764,6 +1956,77 @@ id=netplan-br0
 type=bridge
 interface-name=br0
 
+[ipv4]
+method=auto
+'''})
+        self.assert_networkd({})
+        self.assert_udev(None)
+
+    def test_bridge_params(self):
+        self.generate('''network:
+  version: 2
+  renderer: NetworkManager
+  ethernets:
+    eno1: {}
+    switchport:
+      match:
+        name: enp2s1
+  bridges:
+    br0:
+      interfaces: [eno1, switchport]
+      parameters:
+        ageing-time: 50
+        priority: 1000
+        forward-delay: 12
+        hello-time: 6
+        max-age: 24
+        path-cost:
+          eno1: 70
+        stp: true
+      dhcp4: true''')
+
+        self.assert_nm({'eno1': '''[connection]
+id=netplan-eno1
+type=ethernet
+interface-name=eno1
+slave-type=bridge
+master=br0
+
+[bridge-port]
+path-cost=70
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+''',
+                        'switchport': '''[connection]
+id=netplan-switchport
+type=ethernet
+interface-name=enp2s1
+slave-type=bridge
+master=br0
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+''',
+                        'br0': '''[connection]
+id=netplan-br0
+type=bridge
+interface-name=br0
+
+[bridge]
+ageing-time=50
+priority=1000
+forward-delay=12
+hello-time=6
+max-age=24
+stp=true
+
 [ipv4]
 method=auto
 '''})
@@ -1832,6 +2095,152 @@ id=netplan-bn0
 type=bond
 interface-name=bn0
 
+[ipv4]
+method=auto
+'''})
+        self.assert_networkd({})
+        self.assert_udev(None)
+
+    def test_bond_empty_params(self):
+        self.generate('''network:
+  version: 2
+  renderer: NetworkManager
+  ethernets:
+    eno1: {}
+    switchport:
+      match:
+        name: enp2s1
+  bonds:
+    bn0:
+      interfaces: [eno1, switchport]
+      parameters: {}
+      dhcp4: true''')
+
+        self.assert_nm({'eno1': '''[connection]
+id=netplan-eno1
+type=ethernet
+interface-name=eno1
+slave-type=bond
+master=bn0
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+''',
+                        'switchport': '''[connection]
+id=netplan-switchport
+type=ethernet
+interface-name=enp2s1
+slave-type=bond
+master=bn0
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+''',
+                        'bn0': '''[connection]
+id=netplan-bn0
+type=bond
+interface-name=bn0
+
+[ipv4]
+method=auto
+'''})
+        self.assert_networkd({})
+        self.assert_udev(None)
+
+    def test_bond_with_params(self):
+        self.generate('''network:
+  version: 2
+  renderer: NetworkManager
+  ethernets:
+    eno1: {}
+    switchport:
+      match:
+        name: enp2s1
+  bonds:
+    bn0:
+      interfaces: [eno1, switchport]
+      parameters:
+        mode: 802.1ad
+        lacp-rate: 10
+        mii-monitor-interval: 10
+        min-links: 10
+        up-delay: 10
+        down-delay: 10
+        all-slaves-active: true
+        transmit-hash-policy: none
+        ad-select: none
+        arp-interval: 10
+        arp-validate: all
+        arp-all-targets: all
+        arp-ip-targets:
+          - 10.10.10.10
+          - 20.20.20.20
+        fail-over-mac-policy: none
+        gratuitious-arp: 10
+        packets-per-slave: 10
+        primary-reselect-policy: none
+        resend-igmp: 10
+        learn-packet-interval: 10
+      dhcp4: true''')
+
+        self.assert_nm({'eno1': '''[connection]
+id=netplan-eno1
+type=ethernet
+interface-name=eno1
+slave-type=bond
+master=bn0
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+''',
+                        'switchport': '''[connection]
+id=netplan-switchport
+type=ethernet
+interface-name=enp2s1
+slave-type=bond
+master=bn0
+
+[ethernet]
+wake-on-lan=0
+
+[ipv4]
+method=link-local
+''',
+                        'bn0': '''[connection]
+id=netplan-bn0
+type=bond
+interface-name=bn0
+
+[bond]
+mode=802.1ad
+lacp_rate=10
+miimon=10
+min_links=10
+xmit_hash_policy=none
+ad_select=none
+all_slaves_active=1
+arp_interval=10
+arp_ip_target=10.10.10.10,20.20.20.20
+arp_validate=all
+arp_all_targets=all
+updelay=10
+downdelay=10
+fail_over_mac=none
+num_grat_arp=10
+packets_per_slave=10
+primary_reselect=none
+resend_igmp=10
+lp_interval=10
+
 [ipv4]
 method=auto
 '''})
@@ -2052,17 +2461,6 @@ class TestConfigErrors(TestBase):
         err = self.generate('network:\n  version: 1', expect_fail=True)
         self.assertIn('/a.yaml line 1 column 11: Only version 2 is supported', err)
 
-    def test_duplicate_id(self):
-        err = self.generate('''network:
-  version: 2
-  ethernets:
-    id0:
-      wakeonlan: true
-    id0:
-      wakeonlan: true
-''', expect_fail=True)
-        self.assertIn("Duplicate net definition ID 'id0'", err)
-
     def test_id_redef_type_mismatch(self):
         err = self.generate('''network:
   version: 2
@@ -2116,7 +2514,7 @@ class TestConfigErrors(TestBase):
   bridges:
     br0:
       interfaces: ['foo']''', expect_fail=True)
-        self.assertIn('/a.yaml line 4 column 18: br0: interface foo is not defined\n', err)
+        self.assertIn('/a.yaml line 4 column 19: br0: interface foo is not defined\n', err)
 
     def test_bridge_multiple_assignments(self):
         err = self.generate('''network:
@@ -2385,7 +2783,7 @@ class TestConfigErrors(TestBase):
   version: 2
   vlans:
     ena: {id: 1, link: en1}''', expect_fail=True)
-        self.assertIn('interface en1 is not defined\n', err)
+        self.assertIn('ena: interface en1 is not defined\n', err)
 
     def test_device_bad_route_to(self):
         self.generate('''network:
@@ -2476,6 +2874,167 @@ class TestConfigErrors(TestBase):
         - 192.168.14.2/24
         - 2001:FFfe::1/64''', expect_fail=True)
 
+    def test_bridge_invalid_dev_for_path_cost(self):
+        self.generate('''network:
+  version: 2
+  ethernets:
+    eno1:
+      match:
+        name: eth0
+  bridges:
+    br0:
+      interfaces: [eno1]
+      parameters:
+        path-cost:
+          eth0: 50
+      dhcp4: true''', expect_fail=True)
+
+    def test_bridge_path_cost_already_defined(self):
+        self.generate('''network:
+  version: 2
+  ethernets:
+    eno1:
+      match:
+        name: eth0
+  bridges:
+    br0:
+      interfaces: [eno1]
+      parameters:
+        path-cost:
+          eno1: 50
+          eno1: 40
+      dhcp4: true''', expect_fail=True)
+
+    def test_bridge_invalid_path_cost(self):
+        self.generate('''network:
+  version: 2
+  ethernets:
+    eno1:
+      match:
+        name: eth0
+  bridges:
+    br0:
+      interfaces: [eno1]
+      parameters:
+        path-cost:
+          eno1: aa
+      dhcp4: true''', expect_fail=True)
+
+    def test_bond_invalid_arp_target(self):
+        self.generate('''network:
+  version: 2
+  ethernets:
+    eno1:
+      match:
+        name: eth0
+  bonds:
+    bond0:
+      interfaces: [eno1]
+      parameters:
+        arp-ip-targets:
+          - 2001:dead:beef::1
+      dhcp4: true''', expect_fail=True)
+
+
+class TestForwardDeclaration(TestBase):
+
+    def test_fwdecl_bridge_on_bond(self):
+        self.generate('''network:
+  version: 2
+  bridges:
+    br0:
+      interfaces: ['bond0']
+      dhcp4: true
+  bonds:
+    bond0:
+      interfaces: ['eth0', 'eth1']
+  ethernets:
+    eth0:
+      match:
+        macaddress: 00:01:02:03:04:05
+      set-name: eth0
+    eth1:
+      match:
+        macaddress: 02:01:02:03:04:05
+      set-name: eth1
+''')
+
+        self.assert_networkd({'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n',
+                              'br0.network': ND_DHCP4 % 'br0',
+                              'bond0.netdev': '[NetDev]\nName=bond0\nKind=bond\n',
+                              'bond0.network': '[Match]\nName=bond0\n\n'
+                                               '[Network]\nBridge=br0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n',
+                              'eth0.link': '[Match]\nMACAddress=00:01:02:03:04:05\n\n'
+                                           '[Link]\nName=eth0\nWakeOnLan=off\n',
+                              'eth0.network': '[Match]\nMACAddress=00:01:02:03:04:05\nName=eth0\n\n'
+                                              '[Network]\nBond=bond0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n',
+                              'eth1.link': '[Match]\nMACAddress=02:01:02:03:04:05\n\n'
+                                           '[Link]\nName=eth1\nWakeOnLan=off\n',
+                              'eth1.network': '[Match]\nMACAddress=02:01:02:03:04:05\nName=eth1\n\n'
+                                              '[Network]\nBond=bond0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n'})
+
+    def test_fwdecl_feature_blend(self):
+        self.generate('''network:
+  version: 2
+  vlans:
+    vlan1:
+      link: 'br0'
+      id: 1
+      dhcp4: true
+  bridges:
+    br0:
+      interfaces: ['bond0', 'eth2']
+      parameters:
+        path-cost:
+          eth2: 1000
+          bond0: 8888
+  bonds:
+    bond0:
+      interfaces: ['eth0', 'br1']
+  ethernets:
+    eth0:
+      match:
+        macaddress: 00:01:02:03:04:05
+      set-name: eth0
+  bridges:
+    br1:
+      interfaces: ['eth1']
+  ethernets:
+    eth1:
+      match:
+        macaddress: 02:01:02:03:04:05
+      set-name: eth1
+    eth2:
+      match:
+        name: eth2
+''')
+
+        self.assert_networkd({'vlan1.netdev': '[NetDev]\nName=vlan1\nKind=vlan\n\n'
+                                              '[VLAN]\nId=1\n',
+                              'vlan1.network': ND_DHCP4 % 'vlan1',
+                              'br0.netdev': '[NetDev]\nName=br0\nKind=bridge\n\n'
+                                            '[Bridge]\nSTP=true\n',
+                              'br0.network': '[Match]\nName=br0\n\n'
+                                             '[Network]\nVLAN=vlan1\n',
+                              'bond0.netdev': '[NetDev]\nName=bond0\nKind=bond\n',
+                              'bond0.network': '[Match]\nName=bond0\n\n'
+                                               '[Network]\nBridge=br0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n\n'
+                                               '[Bridge]\nCost=8888\n',
+                              'eth2.network': '[Match]\nName=eth2\n\n'
+                                              '[Network]\nBridge=br0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n\n'
+                                              '[Bridge]\nCost=1000\n',
+                              'br1.netdev': '[NetDev]\nName=br1\nKind=bridge\n',
+                              'br1.network': '[Match]\nName=br1\n\n'
+                                             '[Network]\nBond=bond0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n',
+                              'eth0.link': '[Match]\nMACAddress=00:01:02:03:04:05\n\n'
+                                           '[Link]\nName=eth0\nWakeOnLan=off\n',
+                              'eth0.network': '[Match]\nMACAddress=00:01:02:03:04:05\nName=eth0\n\n'
+                                              '[Network]\nBond=bond0\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n',
+                              'eth1.link': '[Match]\nMACAddress=02:01:02:03:04:05\n\n'
+                                           '[Link]\nName=eth1\nWakeOnLan=off\n',
+                              'eth1.network': '[Match]\nMACAddress=02:01:02:03:04:05\nName=eth1\n\n'
+                                              '[Network]\nBridge=br1\nLinkLocalAddressing=no\nIPv6AcceptRA=no\n'})
+
 
 class TestMerging(TestBase):
     '''multiple *.yaml merging'''
